

/*
	HttpResponse (only native types)
	HttpRequest (only native types)
	HttpConnection (only native types)
	HttpPortListener: requires HttpConnection
	HttpServer: requires HttpResponse, HttpRequest, HttpPortListener
*/

/*****************
Log to console if no logger is specified directly on the window object.
*****************/
	Window.prototype.log=(function(){
		var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
		return function(msg){
			consoleService.logStringMessage(msg);
		};
	})();


/*****************
Fallback since bind isn't natively supported yet.
*****************/
	Function.prototype.bind=Function.prototype.bind || (function(scope,extraArgs){ var target=this; extraArgs=Array.prototype.slice.call(arguments,1); return function(){ return target.apply(scope,extraArgs.concat(Array.prototype.slice.call(arguments))); }; } );


/*****************
{object} HttpResponse ( [{string} data = ""] )
	Constructor:
		[ {string} data = "" ]
			Content to be sent, e.g. "<html><h..."
	Methods:
		{string} toString()
			returns a string that's ready to be sent as a http response, eg "HTTP/1.1 200 ..."
	Properties:
		{string} data [=""]
			Content to be sent, e.g. "<html><h..."
		{string} mime [="text/plain"]
			Mime, eg "text/html"
		{bool} cache [=false]
			Indicates if the content is supposed to be cached
		{string} state [="200 OK"]
			Response code and message, e.g. "404 Not Found"
*****************/
	var HttpResponse=function(data, mime, cache, state){
		log("HttpResponse::ctor");
		if(typeof(data)=="string")
			this.data=data;
		if(typeof(mime)=="string")
			this.mime=mime;
		if(typeof(cache)=="boolean")
			this.cache=cache;
		if(typeof(state)=="string")
			this.state=state;
	};

		HttpResponse.prototype={
			toString:function(){
				log("HttpResponse::toString");
				return "HTTP/1.1 "+this.state+"\r\nContent-Length: "+this.data.length+"\r\nContent-Type: "+this.mime+(this.cache?"":"\r\nCache-Control: no-cache")+"\r\n\r\n"+this.data;
			},
			data:"",
			mime:"text/plain",
			cache:false,
			state:"200 OK"
		};


/*****************
{object} HttpRequest ( {string} rawRequestData , [ {bool} dontStripTrailingSlash = false ] )
	Constructor:
		{string} rawRequestData
			Raw data received
		[{bool} dontStripTrailingSlash = false ]
			Raw data to be sent.
	Methods:
		{string} toString()
			returns a string that's ready to be sent as a http response, eg "HTTP/1.1 200 ..."
	Properties:
		{string} action [=""]
			Action keyword, e.g. "GET"
		{string} url [=""]
			Requested location, e.g. "/foo"
		{object} args [=new Object]
			Arguments passed as part of the location, e.g. a:"b",c:"d" for /index.html?a=b&c=d
*****************/
	var HttpRequest=function(data,dontStripTrailingSlash){
		log("HttpRequest::ctor");
		var parts=/([\S]+) ([^\?]*?)(?: |\?(.*) )/.exec(data);
		this.action=parts[1];
		this.url=dontStripTrailingSlash?parts[2]:parts[2].replace(/.\/$/,"");
		this.args={};
		if(parts[3]){
			var m;
			var comp=/([^=]+)=(.*?)(?:\&|$)/g;
			do{
				m = (comp).exec(parts[3]);
				
				if(m)
					this.args[m[1]]=decodeURIComponent(m[2]);
			}while(m);
		}
	};

		HttpRequest.prototype={
			action:"",
			url:"",
			args:null
		};

/*****************
{object} HttpConnection ( {nsISocketTransport} socketTransport ] )
	Constructor:
		{nsISocketTransport} socketTransport
			Transport opened by the Server Socket
	Methods:
		{void} write()
*****************/
	var HttpConnection=function(aTransport){
		log("HttpConnection::ctor");
		this.__transport=aTransport;
		
		this.__input=aTransport.openInputStream(0,0,0);
		this.__output=aTransport.openOutputStream(0,0,0);
		
		this.__inputReady=this.__inputReady.bind(this);
		this.__outputReady=this.__outputReady.bind(this);
		
		this.__input.asyncWait({onInputStreamReady:this.__inputReady},0, 0, Components.classes["@mozilla.org/thread-manager;1"].getService().mainThread);
		this.__output.asyncWait({onOutputStreamReady:this.__outputReady},0, 0, Components.classes["@mozilla.org/thread-manager;1"].getService().mainThread);
	};

		HttpConnection.prototype={
			onIncoming:null,
			write:function(httpData){
				log("HttpConnection::write");
				this.__writeQueue+=httpData;
				this.__output.asyncWait({onOutputStreamReady:this.__outputReady},0, 0, Components.classes["@mozilla.org/thread-manager;1"].getService().mainThread);
			},
			close:function(kill){
				log("HttpConnection::close");
				this.__closeAfterNextWrite=true;
				if(kill)
					this.__transport.close();
			},
			
			isInputClosed:function(){
				var yes=false;
				try{
					this.__scrinput.available();
				}catch(e if e.result== 0x80470002){
					yes=true;
				}
				log("HttpConnection::isInputClosed , "+yes);
				return yes;
			},
			
			__closeAfterNextWrite:false,
			__transport:null,
			__writeQueue:"",
			__input:null,
			__output:null,
			__scrinput:null,
			
			__outputReady:function(){
				
				if(this.__writeQueue){
					var written=this.__output.write(this.__writeQueue,this.__writeQueue.length);
					this.__output.flush();
					if(written<this.__writeQueue.length){
						log("HttpConnection::__outputReady , rescheduling , written "+written);
						this.__writeQueue=this.__writeQueue.substr(written,this.__writeQueue.length);
						this.__output.asyncWait({onOutputStreamReady:this.__outputReady},0, 0, Components.classes["@mozilla.org/thread-manager;1"].getService().mainThread);
					}else{
						this.__writeQueue="";
						
						if(this.__closeAfterNextWrite){
							log("HttpConnection::__outputReady , closing , written "+written);
							this.__scrinput.close(Components.results.NS_OK);
							this.__input.close(Components.results.NS_OK);
							this.__output.close(Components.results.NS_OK);
							/* Commenting this out (and hoping it will be collected) seems wrong, but closing the transport explicitely aborts any pending transfers and I don't know how to get notified when it's ready to be closed. */
							//this.__transport.close(Components.results.NS_OK);
						}else
							log("HttpConnection::__outputReady , leave open , written "+written);
					}
				}else
					log("HttpConnection::__outputReady");
			},
			__inputReady:function(){
				log("HttpConnection::__inputReady");
				
				if(!this.__scrinput){
					this.__scrinput= Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
					this.__scrinput.init(this.__input);
				}
				
				if(this.isInputClosed())
					return;
				
				var len=this.__scrinput.available();
				if(this.onIncoming && len)
					this.onIncoming(this.__scrinput.read(len));
					
				this.__input.asyncWait({onInputStreamReady:this.__inputReady},0, 0, Components.classes["@mozilla.org/thread-manager;1"].getService().mainThread);
			}
		};


/*****************
	HttpPortListener
*****************/

var HttpPortListener=function(port){
	log("HttpPortListener::ctor");
	var serverSocket = Components.classes["@mozilla.org/network/server-socket;1"] .createInstance(Components.interfaces.nsIServerSocket);
	this.__socket=serverSocket;
	serverSocket.init(port,false,-1);
	serverSocket.asyncListen({
		onSocketAccepted:this.__onSocketAccepted.bind(this),
		onStopListening:this.__onStopListening.bind(this)
	});
	
};

HttpPortListener.prototype={
	onConnect:null,
	close:function(){
		log("HttpPortListener::close");
		if(this.__destroyed) throw("Already destroyed");
		
		this.__socket.close();
	},
	
	__destroyed:false,
	__socket:null,
	__onSocketAccepted:function(aServ,aTransport){
		log("\nHttpPortListener::__onSocketAccepted");
		if(this.onConnect){
			this.onConnect(new HttpConnection(aTransport));
		}else{
			aTransport.close(Components.results.NS_OK);
		}
	},
	__onStopListening:function(){
		log("HttpPortListener::__onStopListening");
		this.__destroyed=true;
	}
};

/*****************
	HttpServer
*****************/

var HttpServer=function(content,port){
	log("HttpServer::ctor");
	if(!port)
		port=80;
	
	this.content=content||{};
	this.__listener=new HttpPortListener(port);
	this.__listener.onConnect=this.__onConnect.bind(this);
};

HttpServer.prototype={
	content:null,
	close:function(){
		log("HttpServer::close");
		this.__listener.close();
	},
	
	__listener:null,
	__onConnect:function(connection){
		log("HttpServer::__onConnect");
		//Run when a new client connects. Schedule __onIncoming to be run when there's data incoming from this client
		connection.onIncoming=this.__onIncoming.bind(this,connection);
	},
	__onIncoming:function(connection,rawData){
		log("HttpServer::__onIncoming");
		var request=new HttpRequest(rawData);
		
		//Get content to be served from content property. If the content is a function, run it.
		if(typeof(this.content[request.url])=="function"){
			var data=this.content[request.url](request);
		}else if(typeof(this.content[request.url])=="undefined"){
			var data=new HttpResponse("Not found");
			data.state="404 Not found";
		}else{
			var data=this.content[request.url];
		}
		
		connection.write(data); //Does not actually write, just schedules writing the next time the output becomes available.
		connection.close(); //Does not actually close, but schedules closing the next time the write buffer is empty.
	}
};